"""
Authentication can follow one of multiple paths e.g. the Authorization Code Flow or the Direct Access Grant.
The flows determine how the ID Token and Access Token are returned to the Relying Party (client).

For details see `Section 3 of OpenID Connect Core 1.0 <https://openid.net/specs/openid-connect-core-1_0.html#Authentication>`_
"""
